"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const Promise = require("bluebird");
const events_1 = require("events");
const zookeeper = require("node-zookeeper-client");
const LocatorException = require("./locatorException");
function deferredLocationRequest() {
    let resolve, reject;
    /* tslint:disable */
    const promise = new Promise(function () {
        resolve = arguments[0];
        reject = arguments[1];
    });
    /* tslint:enable */
    return {
        callback: null,
        promise,
        reject,
        resolve
    };
}
class ZookeeperLocator {
    static getLocatorFactory(parameters) {
        ZookeeperLocator.manager = new ZookeeperLocatorFatory(parameters);
        return ZookeeperLocator.manager.getLocatorForPath;
    }
}
ZookeeperLocator.DefaultLocatorTimeout = 2000;
ZookeeperLocator.DefaultSessionTimeout = 10000;
ZookeeperLocator.DefaultSpinDelay = 1000;
ZookeeperLocator.DefaultRetries = 3;
ZookeeperLocator.DefaultDataExtractor = (data) => {
    let zkJS;
    try {
        zkJS = JSON.parse(data);
    }
    catch (e) {
        return null;
    }
    if (!zkJS.address || !zkJS.port) {
        return null;
    }
    return {
        host: zkJS.address,
        port: zkJS.port
    };
};
exports.ZookeeperLocator = ZookeeperLocator;
/**
 * ZookeeperLocatorFactory manages a zookeeper client connection as well as constructing
 * ZookeeperPathLocatorFactories for each path requests, passing them a handle to the zookeeper
 * client via ZookeeperClientWrapper
 */
class ZookeeperLocatorFatory extends events_1.EventEmitter {
    constructor(parameters) {
        super();
        /**
         * given a path in a zookeeper cluster, return a Locator to resolve to a Location listed in that path
         * @param path
         * @returns {Locator}
         */
        this.getLocatorForPath = (path) => {
            if (typeof path !== "string") {
                throw new TypeError("path must be a string");
            }
            if (path[0] !== "/") {
                path = "/" + path;
            }
            if (!this.pathLocatorFactories[path]) {
                this.pathLocatorFactories[path] = new ZookeeperPathLocatorFactory(this.clientWrapper, path, this, this.dataExtractor, this.locatorTimeout, this.strict);
            }
            return (this.pathLocatorFactories[path].getLocator);
        };
        /**
         * Connect to the zookeeper cluster, rebuilding the client on session expired.
         */
        this.connect = () => {
            this.serverLocator()
                .then((location) => {
                let zookeeperServer = location.host;
                if (location.port) {
                    zookeeperServer = zookeeperServer + ":" + location.port;
                }
                zookeeperServer = zookeeperServer + this.discoveryPath;
                this.client = zookeeper.createClient(zookeeperServer, {
                    retries: this.retries,
                    sessionTimeout: this.sessionTimeout,
                    spinDelay: this.spinDelay
                });
                this.clientWrapper.setClient(this.client);
                let disconnectedHandler = null;
                this.client.on("connected", () => {
                    this.connected = true;
                    this.emit(LocatorException.CODE.CONNECTED);
                    if (disconnectedHandler) {
                        clearTimeout(disconnectedHandler);
                    }
                });
                this.client.on("disconnected", () => {
                    this.connected = false;
                    this.emit(LocatorException.CODE.DISCONNECTED);
                    disconnectedHandler = setTimeout(() => {
                        if (!this.connected) {
                            process.nextTick(() => {
                                this.client.emit("expired");
                            });
                        }
                    }, this.sessionTimeout);
                });
                this.client.on('state', (state) => {
                    this.emit(LocatorException.CODE.STATE_CHANGE, state, this.client.getSessionId());
                });
                this.client.once("expired", () => {
                    this.client.removeAllListeners();
                    this.connected = false;
                    this.client.close();
                    this.emit(LocatorException.CODE.EXPIRED);
                    this.reconnect();
                });
                this.emit(LocatorException.CODE.CONNECTING);
                this.client.connect();
            })
                .catch((err) => {
                this.connected = false;
                this.client.close();
                this.emit(LocatorException.CODE.ZK_LOCATOR_ERROR, err);
                this.connect();
            });
        };
        this.reconnect = () => {
            if (this.spinDelay) {
                setTimeout(() => {
                    this.connect();
                }, this.spinDelay);
            }
            else {
                this.connect();
            }
        };
        const { serverLocator, path, dataExtractor, locatorTimeout, sessionTimeout, spinDelay, retries, strict } = parameters;
        this.connected = false;
        this.serverLocator = serverLocator;
        this.discoveryPath = path;
        this.dataExtractor = dataExtractor ? dataExtractor : ZookeeperLocator.DefaultDataExtractor;
        this.locatorTimeout = locatorTimeout ? locatorTimeout : ZookeeperLocator.DefaultLocatorTimeout;
        this.sessionTimeout = sessionTimeout ? sessionTimeout : ZookeeperLocator.DefaultSessionTimeout;
        this.spinDelay = spinDelay ? spinDelay : ZookeeperLocator.DefaultSpinDelay;
        this.retries = retries ? retries : ZookeeperLocator.DefaultRetries;
        this.strict = strict == null ? true : strict;
        this.clientWrapper = new ZookeeperClientWrapper();
        this.pathLocatorFactories = {};
        // wire up wrapped event emitter to function getLocatorForPath to 'implement' the Locator interface
        // ugly, but this whole shared event emitter business sort of is :(
        // also, how only the zookeeper locator implements this :(
        Object.keys(events_1.EventEmitter.prototype).forEach((fnName) => (this.getLocatorForPath)[fnName] = this[fnName]);
        this.connect();
    }
}
/**
 * ZookeeperPathLocatorFactory provides Locators for a list of hosts stored in a path in a zookeeper cluster
 */
class ZookeeperPathLocatorFactory {
    constructor(clientWrapper, path, emitter, dataExtractor, locatorTimeout, strict) {
        /**
         * get a location from zookeeper for this path. 'implements' the EmitterLocator interface
         * @returns {Bluebird<R>}
         */
        this.getLocator = () => {
            const locator = deferredLocationRequest();
            if (this.stale) {
                this.setLocationPool(null);
                this.clientWrapper.client.exists(this.path, this.onLocationPathExists);
            }
            if (this.locationPool) {
                this.resolveLocationRequest(locator);
            }
            else {
                this.locationRequestQueue.push(locator);
            }
            if (this.locatorTimeout) {
                setTimeout((() => {
                    if (locator.promise.isPending()) {
                        locator.reject(new Error("ZOOKEEPER_TIMEOUT"));
                    }
                }), this.locatorTimeout);
            }
            return locator.promise;
        };
        /**
         * zookeeper.client.getChildren response handler.  upon fetching the path we get a list of children nodes,
         * call getInfo on them, and store them in the location pool and kick of the location request processing queue
         * @param err
         * @param children
         */
        this.onGetLocations = (err, children) => {
            if (err) {
                this.emitter.emit(LocatorException.CODE.FAILED_TO_GET_CHILDREN, this.path, err);
                this.setLocationPool([]);
                this.processQueue();
                return;
            }
            const promises = children.map((child) => {
                const locator = deferredLocationRequest();
                this.clientWrapper.client.getData(this.path + "/" + child, (_err, data) => {
                    if (_err) {
                        if (_err.getCode() !== zookeeper.Exception.NO_NODE) {
                            this.emitter.emit(LocatorException.CODE.FAILED_TO_GET_CHILD_INFO, this.path, _err);
                        }
                        locator.resolve(null);
                        return;
                    }
                    locator.resolve(this.dataExtractor(data.toString("utf8")));
                    return;
                });
                return locator.promise;
            });
            Promise.all(promises)
                .then((newPool) => {
                this.setLocationPool(newPool.filter(Boolean));
                this.emitter.emit(LocatorException.CODE.NEW_POOL, this.path, this.locationPool);
                this.processQueue();
            }).catch((allErr) => {
                this.emitter.emit(LocatorException.CODE.FAILED_TO_GET_CHILDREN, this.path, allErr);
                this.setLocationPool([]);
                this.processQueue();
                return;
            });
        };
        /**
         * changed handler for zookeeper.client.getChildren method. when the list of locations change, fetch
         * the list of children and forward the event
         * @param event
         */
        this.onLocationsChange = (event) => {
            this.emitter.emit(LocatorException.CODE.CHILDREN_CHANGED, this.path, event);
            this.clientWrapper.client.getChildren(this.path, this.onLocationsChange, this.onGetLocations);
        };
        /**
         * zookeeper.client.exists handler to check and see that a location path exists in the zookeeper cluster
         * @param error zookeeper client error
         * @param stat zookeeper.Stat object with information about the zookeeper location path
         */
        this.onLocationPathExists = (error, stat) => {
            if (stat) {
                this.stale = false;
                this.emitter.emit(LocatorException.CODE.PATH_FOUND, this.path);
                this.clientWrapper.client.getChildren(this.path, this.onLocationsChange, this.onGetLocations);
            }
            else {
                this.stale = true;
                this.emitter.emit(LocatorException.CODE.PATH_NOT_FOUND, this.path);
                this.setLocationPool([]);
                this.processQueue();
            }
        };
        this.clientWrapper = clientWrapper;
        this.path = path;
        this.emitter = emitter;
        this.dataExtractor = dataExtractor;
        this.locatorTimeout = locatorTimeout;
        this.strict = strict;
        this.next = -1;
        this.locationPool = null;
        this.cachedLocationPool = null;
        this.locationRequestQueue = [];
        this.stale = false;
        clientWrapper.on('NEW_CLIENT', () => {
            this.setLocationPool(null);
            clientWrapper.client.exists(path, this.onLocationPathExists);
        });
        if (clientWrapper.client) {
            clientWrapper.client.exists(path, this.onLocationPathExists);
        }
    }
    /**
     * update the list of locations available at this path
     * @param newPool
     */
    setLocationPool(newPool) {
        this.locationPool = newPool;
        if (!this.cachedLocationPool || (newPool && newPool.length)) {
            this.cachedLocationPool = newPool;
        }
        return;
    }
    /**
     * get the list of locations if it exists or in strict mode, otherwise try the cached
     * list of locations
     * @returns {Location[]}
     */
    getLocationPool() {
        if (this.strict || (this.locationPool && this.locationPool.length)) {
            return this.locationPool;
        }
        if (this.cachedLocationPool && this.cachedLocationPool.length) {
            return this.cachedLocationPool;
        }
        return this.locationPool;
    }
    /**
     * resolve a location request with a location, or reject if the pool is empty or still loading
     * @param deferred
     */
    resolveLocationRequest(deferred) {
        const chosenPool = this.getLocationPool();
        if (!chosenPool) {
            return deferred.reject(Error("get next called on loading pool"));
        }
        if (chosenPool.length) {
            this.next++;
            return deferred.resolve(chosenPool[this.next % chosenPool.length]);
        }
        else {
            return deferred.reject(new Error(LocatorException.CODE.EMPTY_POOL));
        }
    }
    /**
     * process the queue of location requests for this zookeeper path
     */
    processQueue() {
        while (this.locationRequestQueue.length) {
            this.resolveLocationRequest(this.locationRequestQueue.shift());
        }
    }
}
/**
 * ZookeeperClientWrapper is used to supply a zookeeper client
 * connection to each ZookeeperPathLocatorFactory by the ZookeeperLocatorFactory,
 * without having to update the references to the zookeeper client when it is replaced
 * due to expired session
 */
class ZookeeperClientWrapper extends events_1.EventEmitter {
    constructor() {
        super();
        this.client = null;
    }
    /**
     * Set the zookeeper client and emit an event in case anyone is watching
     * @param initialized client zookeeper client
     */
    setClient(client) {
        this.client = client;
        this.emit('NEW_CLIENT');
    }
}
//# sourceMappingURL=zookeeper.js.map